/* rects - rectangle routines
**
** Copyright (C) 1992 by Jef Poskanzer
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
*/

#ifndef lint
static char rcsid[] =
    "@(#) $Header: Rects.c,v 1.1 92/03/11 19:40:31 jef Exp $";
#endif

#include <X11/Xlib.h>
#include <stdio.h>
#include "Rects.h"


/* Definitions. */

static int num_exposerects, max_exposerects;
static XRectangle* exposerects;


/* Routines. */

/* Initialize the Rects package. */
void
RectsInit()
    {
    num_exposerects = max_exposerects = 0;
    }

/* Queue up a rectangle for repainting - better than sending an Expose event. */
void
RectsInvalidate( rectP, merge )
    XRectangle* rectP;
    Bool merge;
    {
    int i;

    if ( merge )
	{
	/* Check if this rectangle intersects an existing one. */
	for ( i = 0; i < num_exposerects; ++i )
	    {
	    if ( RectsTouch( &exposerects[i], rectP ) )
		{
		RectsUnion( &exposerects[i], rectP );
		return;
		}
	    }
	}

    /* Nope, add a new XRectangle. */
    if ( num_exposerects == max_exposerects )
	{
	if ( max_exposerects == 0 )
	    {
	    max_exposerects = 32;	/* non-critical parameter */
	    exposerects = (XRectangle*) malloc(
		(unsigned) ( max_exposerects * sizeof(XRectangle) ) );
	    }
	else
	    {
	    max_exposerects *= 2;
	    exposerects = (XRectangle*) realloc(
		(char*) exposerects,
		(unsigned) ( max_exposerects * sizeof(XRectangle) ) );
	    }
	if ( exposerects == (XRectangle*) 0 )
	    {
	    (void) fprintf( stderr, "RectsInvalidate: out of memory\n" );
	    exit( 1 );
	    }
	}
    exposerects[num_exposerects] = *rectP;
    ++num_exposerects;
    }

/* Return a rectangle to be repainted, if there are any. */
Bool
RectsNextInvalid( rectP )
    XRectangle* rectP;
    {
    if ( num_exposerects == 0 )
	return False;
    --num_exposerects;
    *rectP = exposerects[num_exposerects];
    return True;
    }

/* Make rect1 the union of rect1 and rect2. */
void
RectsUnion( rect1P, rect2P )
    XRectangle* rect1P;
    XRectangle* rect2P;
    {
    if ( rect2P->x + rect2P->width > rect1P->x + rect1P->width )
	rect1P->width = rect2P->x + rect2P->width - rect1P->x;
    if ( rect2P->y + rect2P->height > rect1P->y + rect1P->height )
	rect1P->height = rect2P->y + rect2P->height - rect1P->y;
    if ( rect2P->x < rect1P->x )
	{
	rect1P->width = rect1P->x + rect1P->width - rect2P->x;
	rect1P->x = rect2P->x;
	}
    if ( rect2P->y < rect1P->y )
	{
	rect1P->height = rect1P->y + rect1P->height - rect2P->y;
	rect1P->y = rect2P->y;
	}
    }

/* Make rect1 the intersection of rect1 and rect2. */
void
RectsIntersect( rect1P, rect2P )
    XRectangle* rect1P;
    XRectangle* rect2P;
    {
    if ( ! RectsTouch( rect1P, rect2P ) )
	{
	rect1P->width = rect1P->height = 0;
	return;
	}
    if ( rect2P->x + rect2P->width < rect1P->x + rect1P->width )
	rect1P->width = rect2P->x + rect2P->width - rect1P->x;
    if ( rect2P->y + rect2P->height < rect1P->y + rect1P->height )
	rect1P->height = rect2P->y + rect2P->height - rect1P->y;
    if ( rect2P->x > rect1P->x )
	{
	rect1P->width = rect1P->x + rect1P->width - rect2P->x;
	rect1P->x = rect2P->x;
	}
    if ( rect2P->y > rect1P->y )
	{
	rect1P->height = rect1P->y + rect1P->height - rect2P->y;
	rect1P->y = rect2P->y;
	}
    }

/* Test whether rect1 and rect2 intersect. */
Bool
RectsTouch( rect1P, rect2P )
    XRectangle* rect1P;
    XRectangle* rect2P;
    {
    if ( rect1P->x + rect1P->width <= rect2P->x ||
         rect1P->y + rect1P->height <= rect2P->y ||
         rect2P->x + rect2P->width <= rect1P->x ||
         rect2P->y + rect2P->height <= rect1P->y )
	return False;
    return True;
    }

/* Test whether rect1 is inside rect2. */
Bool
RectsInside( rect1P, rect2P )
    XRectangle* rect1P;
    XRectangle* rect2P;
    {
    if ( rect1P->x >= rect2P->x &&
	 rect1P->x + rect1P->width <= rect2P->x + rect2P->width &&
         rect1P->y >= rect2P->y &&
	 rect1P->y + rect1P->height <= rect2P->y + rect2P->height )
	return True;
    return False;
    }
